#include <fstream>
#include <map>

#include "MCFgen.hpp"  // die()
#include "cis_element_model.h"

void cis_element_submodel::counts2probs(double pseudocount)
{
  for (unsigned int k = 0u; k < matrix.size(); ++k) {
    double tot = 0.0;
    for (unsigned int i = 0u; i < matrix[k].size(); ++i) {
      matrix[k][i] += pseudocount;
      tot += matrix[k][i];
    }
    for (unsigned int i = 0u; i < matrix[k].size(); ++i)
      matrix[k][i] /= tot;
  }
}

void cis_element_submodel::set_undefined_weights(double default_weight)
{
  if (plus_strand_weight == -1.0)
    plus_strand_weight = default_weight;
  if (minus_strand_weight == -1.0)
    minus_strand_weight = default_weight;
}

void cis_element_model::set_undefined_weights(void)
{
  for (unsigned int s = 0u; s < submodels.size(); ++s)
    submodels[s].set_undefined_weights(1.0 / submodels.size());
}

void cis_element_model::normalize_weights_within(void)
{
  double tot = 0.0;

  for (unsigned int s = 0u; s < submodels.size(); ++s) {
    tot += submodels[s].plus_strand_weight;
    tot += submodels[s].minus_strand_weight;
  }

  for (unsigned int s = 0u; s < submodels.size(); ++s) {
    submodels[s].plus_strand_weight /= tot;
    submodels[s].minus_strand_weight /= tot;
  }
}

void cis_element_model::counts2probs(double pseudocount)
{
  for (unsigned int s = 0u; s < submodels.size(); ++s)
    submodels[s].counts2probs(pseudocount);
}

istream & operator >> (istream & strm, cis_element_submodel & s)
{
  char c;
  string label;
  double plus_strand_weight = -1.0;
  double minus_strand_weight = -1.0;
  vector<int> transparency_mask;
  vector<vector<double> > matrix;

  while(strm >> c) {  // read next non-whitespace character
    switch (c) {
    case 'L':  // a "label"
      strm >> ws;  // ignore whitespace
      getline(strm, label);
      break;
    case '+':  // a weight for the + strand
      strm >> plus_strand_weight;
      break;
    case '-':  // a weight for the - strand
      strm >> minus_strand_weight;
      break;
    case '>':  // a title line
      if(strm.unget()) {  // leave it for the next read
	s = cis_element_submodel(label, plus_strand_weight, minus_strand_weight, transparency_mask, matrix);
	return strm;  // we're done
      }
      break;
    case 'n':
    case 'N':
    case 'b':
    case 'B':  // a "transparent" (= background) column
      transparency_mask.push_back(1);
      // need to push back a dummy column, so things don't break
      matrix.push_back(vector<double>(4, 1.0));
      break;
    default:  // must be counts for A, C, G, T
      strm.unget();
      transparency_mask.push_back(0);
      matrix.push_back(vector<double>(0));
      for (int i = 0; i < 4; ++i) {
	double count;
	strm >> count;
	matrix.back().push_back(count);
      }
    }
    if (!strm)  // check for read error (don't clear EOF here)
      return strm;
  }

  // if reached EOF, clear the stream state: (so it will happily read nothing)
  if (strm.eof() && !strm.bad())
    strm.clear();

  if (strm)  // assign to s only if no read error occurred
    s = cis_element_submodel(label, plus_strand_weight, minus_strand_weight, transparency_mask, matrix);

  return strm;
}

// read all cis-elements from the file
void cis_element_model::read_file(const string & filename, vector<cis_element_model> & models)
{
  ifstream file(filename.c_str());
  if (!file) mcf::die("Sorry, couldn't open file " + filename);

  char c;

  // map to associate model names with indices of the models vector:
  map<string, unsigned int> model_indices;

  while (file >> c >> ws) {  // read a character & ignore whitespace
    if (c != '>')  // the first character must be '>'
      mcf::die("Sorry, couldn't understand the matrix file " + filename);
    string name;
    getline(file, name);
    cis_element_submodel s;  // hope it gets reset each time through the loop!
    if (!(file >> s) || s.matrix.size() == 0)
      mcf::die("Sorry, couldn't understand the matrix file " + filename);
    map<string, unsigned int>::iterator pos = model_indices.find(name);
    if (pos == model_indices.end()) {
      model_indices[name] = models.size();
      models.push_back(cis_element_model(name, vector<cis_element_submodel>(1, s)));
    } else
      models[pos->second].submodels.push_back(s);
  }

  for (unsigned int m = 0u; m < models.size(); ++m)
    models[m].set_undefined_weights();
}

// just get the cis-element with the specified title
void cis_element_model::read_one(const string & filename, const string & title, 
				    cis_element_model & model)
{
  ifstream file(filename.c_str());
  if (!file) mcf::die("Sorry, couldn't open file " + filename);

  vector<cis_element_submodel> subs;
  char c;

  while (file >> c >> ws) {  // read a character & ignore whitespace
    if (c != '>')  // the first character must be '>'
      mcf::die("Sorry, couldn't understand the matrix file " + filename);
    string name;
    getline(file, name);
    cis_element_submodel s;  // hope it gets reset each time through the loop!
    if (!(file >> s) || s.matrix.size() == 0)
      mcf::die("Sorry, couldn't understand the matrix file " + filename);
    if (name == title)
      subs.push_back(s);
  }

  if (subs.size() == 0)
    mcf::die("No cis-element \"" + title + "\" in " + filename);
  model = cis_element_model(title, subs);

  model.set_undefined_weights();
}
