/*
possum: scan individual PSSMs against sequences
now brought into the C++ age!
allows a cis-element to be represented by multiple matrices
*/

#include <algorithm>  // reverse()
#include <fstream>
#include <iomanip>  // setw()
#include <iostream>
#include <numeric>  // accumulate()
#include <string>
#include <vector>
#include <cmath>  // log(), exp(), log10()

#include "misc.h"
#include "MCFgen.hpp"
#include "MCFbio.hpp"
#include "args.hpp"
#include "cis_element_model.h"

// how many digits needed to write the number in base 10?
// what's the coolest way to do this?
unsigned num_digits(unsigned x)
{
  unsigned n = 1;
  while (x /= 10) ++n;
  return n;
}

// DNA-to-number translator that only recognizes uppercase
inline unsigned nolower_translator(char c)
{
  switch(c) {
  case 'A':
    return 0u;
  case 'C':
    return 1u;
  case 'G':
    return 2u;
  case 'T':
    return 3u;
  default:
    return 4u;
  }
}

using namespace std;

struct matrix {
  string label;
  char strand;  // + or - or p
  vector<vector<double> > prob_matrix;
  vector<int> transparency_mask;  // indicates positions with emission probs = background
  unsigned int width;
  double weight;
  double threshold; // = prob_ratio_threhold / weight

  // constructor:
  matrix(const cis_element_submodel & c, char s, double w, const string & alt_name);

  // convert a cis-element model to a vector of matrices:
  static void convert_cis_element(cis_element_model & raw, vector<matrix> & cooked,
				  double pseudocount, double prob_ratio_threshold);
};

matrix::matrix(const cis_element_submodel & c, char s, double w, const string & alt_name)
  : label(c.label), strand(s), prob_matrix(c.matrix),
    transparency_mask(c.transparency_mask), width(c.matrix.size()), weight(w)
{
  if (label.empty())
    label = alt_name;

  if (strand == '-') {
    reverse(transparency_mask.begin(), transparency_mask.end());
    reverse_complement(prob_matrix);
  }

  for (unsigned int k = 0u; k < prob_matrix.size(); ++k)
    prob_matrix[k].push_back(0);  // prob(masked base) = 0
}

namespace possum {
  vector<matrix> matrices;
  vector<uint> seq;
  unsigned int seqlen;
  unsigned int window;
  unsigned int memory;  // = maximum matrix width
  vector<unsigned int> local_base_counts;
  unsigned int local_base_tot;
  unsigned int n;  // holds current position in sequence

  // field widths for printing results prettily:
  uint label_field_width;
  int position_field_width;
  int seq_field_width;

  void init_background(void);
  void update_background(void);
  void print_hit(ostream & output, const matrix & m, double score);
  void scan_seq(ostream & output);
  void init_misc(unsigned int w);
  void read_cis_elements(const string & filename, const string & title,
			 double pseudocount, double prob_ratio_threshold);
}

void possum::init_background(void)
{
  local_base_counts.assign(4, 0u);

  for (unsigned int n = 0u; n < window && n < seqlen; ++n) {
    uint i = seq[n];
    if (i != 4)
      ++local_base_counts[i];
  }

  local_base_tot = accumulate(local_base_counts.begin(), local_base_counts.end(), 0u);

  local_base_counts.push_back(1u);  // for masked bases
}

void possum::update_background(void)
{
  if (n > window) {
    uint i = seq[n - window - 1];
    if (i != 4) {
      --local_base_counts[i];
      --local_base_tot;
    }
  }

  if (n + window < seqlen) {  // beware of unsigned ints !!!
    uint i = seq[n + window];
    if (i != 4) {
      ++local_base_counts[i];
      ++local_base_tot;
    }
  }
}

void possum::print_hit(ostream & output, const matrix & m, double score)
{
  unsigned int start = (n + 1) - m.width;

  string s;
  for (unsigned int k = start; k <= n; ++k)
    s += mcf::number_to_DNA(seq[k]);

  output.setf(ios::left);  // left adjustment
  output << setw(label_field_width) << m.label.c_str();  // why need c_str ???
  output.setf(ios::right);  // right adjustment
  output << ' ' << setw(position_field_width) << (start + 1)
	 << " to " << setw(position_field_width) << (n + 1)
	 << "  " << m.strand;
  output.setf(ios::left);  // left adjustment
  output << "  " << setw(seq_field_width) << s.c_str();
  output << "  " << score << '\n';
}

// the main algorithm
void possum::scan_seq(ostream & output)
{
  init_background();
  vector<double> bg_probs(memory);

  for (n = 0; n < seq.size(); ++n) {

    if (args::even_bg)
      bg_probs[n % memory] = 0.25;
    else {
      update_background();
      bg_probs[n % memory] = local_base_counts[seq[n]]
	/ double(local_base_tot);
    }

    for (unsigned int m = 0u; m < matrices.size(); ++m) {
      const matrix & mat = matrices[m];
      unsigned int width = mat.width;
      if (width > n + 1)
	continue;

      double score = 1.0;
      for (unsigned int k = 0u, pos = (n + 1) - width; k < width; ++k, ++pos)
	if (mat.transparency_mask[k] == 0)
	  score *= mat.prob_matrix[k][seq[pos]] / bg_probs[pos % memory];

      if (score > mat.threshold)
	print_hit(output, mat, log(score * mat.weight));
    }

  }

}

void possum::init_misc(unsigned int w)
{
  possum::window = w;

  // set variable 'memory' to maximum matrix width:
  memory = 0u;
  for (unsigned int m = 0u; m < matrices.size(); ++m)
    if (matrices[m].width > memory)
      memory = matrices[m].width;

  // set label_field_width to length of longest matrix label:
  label_field_width = 0;
  for (unsigned int m = 0u; m < matrices.size(); ++m)
    if (matrices[m].label.size() > label_field_width)
      label_field_width = matrices[m].label.size();

  // set seq_field_width to maximum matrix width:
  seq_field_width = memory;
}

// rescale weights so that the maximum is 1
// also penalize palindromes by a factor of 2
// a bit sketchy - what's the best way to do this?
void normalize_weights_my_way(vector<matrix> & mats)
{
  double max = 0.0;

  for (unsigned int m = 0u; m < mats.size(); ++m) {
    if (mats[m].strand == 'p')
      mats[m].weight /= 2;  // penalty for palindromes
    if (mats[m].weight > max)
      max = mats[m].weight;
  }

  for (unsigned int m = 0u; m < mats.size(); ++m)
    mats[m].weight /= max;
}

bool is_palindrome(const cis_element_submodel & s)
{
  return (mcf::is_reverse(s.transparency_mask.begin(),
			  s.transparency_mask.end())
	  && is_reverse_complement(s.matrix));
}

// convert a cis-element model from "raw" to "cooked" format
// assumes "cooked" is empty
void matrix::convert_cis_element(cis_element_model & raw, vector<matrix> & cooked,
				 double pseudocount, double prob_ratio_threshold)
{
  raw.counts2probs(pseudocount);  // convert counts to probabilities

  for (unsigned int s = 0u; s < raw.submodels.size(); ++s) {
    const cis_element_submodel & sub = raw.submodels[s];

    if (is_palindrome(sub)) {
      double weight = sub.plus_strand_weight + sub.minus_strand_weight;
      if (weight > 0.0)
	cooked.push_back(matrix(sub, 'p', weight, raw.name));
    } else {
      if (sub.plus_strand_weight > 0.0)
	cooked.push_back(matrix(sub, '+', sub.plus_strand_weight, raw.name));
      if (sub.minus_strand_weight > 0.0)
	cooked.push_back(matrix(sub, '-', sub.minus_strand_weight, raw.name));
    }
  }

  normalize_weights_my_way(cooked);

  for (unsigned int m = 0u; m < cooked.size(); ++m)
    cooked[m].threshold = prob_ratio_threshold / cooked[m].weight;
}

// read cis-element models from a file
// & convert them to the format we want
void possum::read_cis_elements(const string & filename, const string & title, 
			       double pseudocount, double prob_ratio_threshold)
{
  if (title.empty()) {  // user didn't specify a cis-element, so read them all in

    vector<cis_element_model> raw_vec;
    cis_element_model::read_file(filename, raw_vec);

    for (unsigned int r = 0u; r < raw_vec.size(); ++r) {
      vector<matrix> new_mats;  // assume it gets cleared each iteration
      matrix::convert_cis_element(raw_vec[r], new_mats, pseudocount, prob_ratio_threshold);
      for (unsigned int m = 0u; m < new_mats.size(); ++m)
	matrices.push_back(new_mats[m]);
    }

  } else {  // just read in the cis-element with the specified title

    cis_element_model raw;
    cis_element_model::read_one(filename, title, raw);
    matrix::convert_cis_element(raw, matrices, pseudocount, prob_ratio_threshold);

  }
}

// return either an open file stream for writing output,
// or cout
ostream & get_output_stream(const string & filename, ofstream & file_stream)
{
  if (filename.empty())
    return cout;
  file_stream.open(filename.c_str());
  if (file_stream)
    return file_stream;
  cerr << "Sorry, couldn't open file " << filename << ", reverting to stdout" << endl;
  return cout;
}

// I'll probably move this to a library
inline std::ifstream & open_or_die(const std::string & filename, std::ifstream & strm)
{
  strm.open(filename.c_str());
  if (!strm) mcf::die("Sorry, couldn't open " + filename);
  return strm;
}

int main(int argc, char **argv)
{
  // parse the command line:
  args::parse(argc, argv);

  // read matrices:
  possum::read_cis_elements(args::matfile, args::mat_title,
			    args::pseudo, exp(args::score_thresh));

  possum::init_misc(args::bg_range);

  // open file to read sequences from:
  ifstream file;
  istream & in = args::seqfile.empty() ? (istream &)cin : open_or_die(args::seqfile, file);

  // decide where to send output to:
  ofstream output_file;
  ostream & output = get_output_stream(args::outfile, output_file);

  output << "POSSUM\n\n";
  args::print(output);
  output << endl;

  output.precision(3);  // precision for floats
  output.setf(ios::showpoint);  // show trailing zeros

  // read sequences one at a time, and scan them:
  unsigned (*translator)(char) = (args::mask_lower ? nolower_translator : mcf::DNA_to_number);
  string seqname;
  while (mcf::get_fasta(in, possum::seq, seqname, translator)) {
    output << '>' << seqname << '\n' << endl;
    possum::seqlen = possum::seq.size();
    possum::position_field_width = num_digits(possum::seqlen);
    possum::scan_seq(output);
    output << endl;
    possum::seq.clear();
  }
}
